#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.polyflow import ForConfig, IfConfig


class Sections:
    VERSION = "version"
    KIND = "kind"
    NAME = "name"
    DESCRIPTION = "description"
    TAGS = "tags"
    PROFILE = "profile"
    QUEUE = "queue"
    CACHE = "cache"
    PLUGINS = "plugins"
    TERMINATION = "termination"
    INPUTS = "inputs"
    OUTPUTS = "outputs"
    PARAMS = "params"
    CONNECTIONS = "connections"
    RUN = "run"
    RUN_PATCH = "runPatch"
    MATRIX = "matrix"
    OPERATIONS = "operations"
    COMPONENTS = "components"
    SCHEDULE = "schedule"
    DEPENDENCIES = "dependencies"
    TRIGGER = "trigger"
    CONDITIONS = "conditions"
    SKIP_ON_UPSTREAM_SKIP = "skipOnUpstreamSkip"
    HUB_REF = "hubRef"
    DAG_REF = "dagRef"
    PATH_REF = "pathRef"
    URL_REF = "urlRef"
    COMPONENT = "component"

    SECTIONS = (
        VERSION,
        KIND,
        NAME,
        DESCRIPTION,
        TAGS,
        PARAMS,
        PROFILE,
        CACHE,
        PLUGINS,
        TERMINATION,
        CONNECTIONS,
        MATRIX,
        OPERATIONS,
        SCHEDULE,
        DEPENDENCIES,
        TRIGGER,
        CONDITIONS,
        SKIP_ON_UPSTREAM_SKIP,
        HUB_REF,
        DAG_REF,
        PATH_REF,
        URL_REF,
        COMPONENT,
        INPUTS,
        OUTPUTS,
        RUN,
        RUN_PATCH,
    )

    PARSING_SECTIONS = (PROFILE, QUEUE, CACHE, CONNECTIONS, PLUGINS, TERMINATION)
    OP_PARSING_SECTIONS = (OPERATIONS, SCHEDULE, DEPENDENCIES, TRIGGER, CONDITIONS)

    REQUIRED_SECTIONS = (VERSION, KIND)

    OPERATORS = {ForConfig.IDENTIFIER: ForConfig, IfConfig.IDENTIFIER: IfConfig}



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import numpy as np
import os
import tempfile

import pytest

from bokeh.plotting import figure
from PIL import Image
from plotly import figure_factory
from tests.utils import BaseTestCase, tensor_np

from polyaxon.polyboard.processors.events_processors import (
    audio,
    bokeh_chart,
    convert_to_HWC,
    histogram,
    image,
    image_boxes,
    plotly_chart,
    prepare_video,
    video,
)


@pytest.mark.tracking_mark
class TestEventValues(BaseTestCase):
    def setUp(self):
        self.run_path = tempfile.mkdtemp()
        self.asset_path = self.run_path + "/asset"

    def test_uint8_image(self):
        """Tests that uint8 image (pixel values in [0, 255]) is not changed"""
        assert os.path.exists(self.asset_path) is False
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(3, 32, 32), dtype=np.uint8),
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_float32_image(self):
        """Tests that float32 image (pixel values in [0, 1]) are scaled correctly to [0, 255]"""
        assert os.path.exists(self.asset_path) is False
        event = image(asset_path=self.asset_path, data=tensor_np(shape=(3, 32, 32)))
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_float_1_converts_to_uint8_255(self):
        assert os.path.exists(self.asset_path) is False
        green_uint8 = np.array([[[0, 255, 0]]], dtype="uint8")
        green_float32 = np.array([[[0, 1, 0]]], dtype="float32")

        a = image(asset_path=self.run_path + "/asset1", data=green_uint8)
        b = image(asset_path=self.run_path + "/asset2", data=green_float32)
        self.assertEqual(
            Image.open(io.BytesIO(open(a.path, "br").read())),
            Image.open(io.BytesIO(open(b.path, "br").read())),
        )

    def test_list_input(self):
        with pytest.raises(Exception):
            histogram("dummy", [1, 3, 4, 5, 6], "tensorflow")

    def test_empty_input(self):
        print("expect error here:")
        with pytest.raises(Exception):
            histogram("dummy", np.ndarray(0), "tensorflow")

    def test_image_with_boxes(self):
        event = image_boxes(
            asset_path=self.asset_path,
            tensor_image=tensor_np(shape=(3, 32, 32)),
            tensor_boxes=np.array([[10, 10, 40, 40]]),
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_with_one_channel(self):
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(1, 8, 8)),
            dataformats="CHW",
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_with_four_channel(self):
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(4, 8, 8)),
            dataformats="CHW",
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_with_one_channel_batched(self):
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(2, 1, 8, 8)),
            dataformats="NCHW",
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_with_3_channel_batched(self):
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(2, 3, 8, 8)),
            dataformats="NCHW",
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_with_four_channel_batched(self):
        event = image(
            asset_path=self.asset_path,
            data=tensor_np(shape=(2, 4, 8, 8)),
            dataformats="NCHW",
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_image_without_channel(self):
        event = image(
            asset_path=self.asset_path, data=tensor_np(shape=(8, 8)), dataformats="HW"
        )
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_video(self):
        asset_path = self.asset_path + ".gif"
        event = video(asset_path=asset_path, tensor=tensor_np(shape=(4, 3, 1, 8, 8)))
        assert event.path == asset_path
        assert os.path.exists(asset_path) is True
        event = video(
            asset_path=asset_path, tensor=tensor_np(shape=(16, 48, 1, 28, 28))
        )
        assert event.path == asset_path
        assert os.path.exists(asset_path) is True
        event = video(asset_path=asset_path, tensor=tensor_np(shape=(20, 7, 1, 8, 8)))
        assert event.path == asset_path
        assert os.path.exists(asset_path) is True

    def test_audio(self):
        event = audio(asset_path=self.asset_path, tensor=tensor_np(shape=(42,)))
        assert event.path == self.asset_path
        assert os.path.exists(self.asset_path) is True

    def test_histogram_auto(self):
        event = histogram(values=tensor_np(shape=(1024,)), bins="auto", max_bins=5)
        assert event.values is not None
        assert event.counts is not None

    def test_histogram(self):
        event = histogram(values=tensor_np(shape=(1024,)), bins="fd", max_bins=5)
        assert event.values is not None
        assert event.counts is not None

    def test_histogram_doane(self):
        event = histogram(tensor_np(shape=(1024,)), bins="doane", max_bins=5)
        assert event.values is not None
        assert event.counts is not None

    def test_to_HWC(self):  # noqa
        np.random.seed(1)
        test_image = np.random.randint(0, 256, size=(3, 32, 32), dtype=np.uint8)
        converted = convert_to_HWC(test_image, "chw")
        assert converted.shape == (32, 32, 3)
        test_image = np.random.randint(0, 256, size=(16, 3, 32, 32), dtype=np.uint8)
        converted = convert_to_HWC(test_image, "nchw")
        assert converted.shape == (64, 256, 3)
        test_image = np.random.randint(0, 256, size=(32, 32), dtype=np.uint8)
        converted = convert_to_HWC(test_image, "hw")
        assert converted.shape == (32, 32, 3)

    def test_prepare_video(self):
        # at each timestep the sum over all other dimensions of the video should stay the same
        np.random.seed(1)
        video_before = np.random.random((4, 10, 3, 20, 20))
        video_after = prepare_video(np.copy(video_before))
        video_before = np.swapaxes(video_before, 0, 1)
        video_before = np.reshape(video_before, newshape=(10, -1))
        video_after = np.reshape(video_after, newshape=(10, -1))
        np.testing.assert_array_almost_equal(
            np.sum(video_before, axis=1), np.sum(video_after, axis=1)
        )

    def test_bokeh_chart(self):
        # prepare some data
        x = [1, 2, 3, 4, 5]
        y = [6, 7, 2, 4, 5]

        # create a new plot with a title and axis labels
        p = figure(title="simple line example", x_axis_label="x", y_axis_label="y")

        # add a line renderer with legend and line thickness
        p.line(x, y, line_width=2)

        # show the results
        event = bokeh_chart(p)
        assert isinstance(event.figure, dict)

    def test_plotly_chart(self):
        x1 = np.random.randn(200) - 2
        x2 = np.random.randn(200)
        x3 = np.random.randn(200) + 2
        hist_data = [x1, x2, x3]
        group_labels = ["Group 1", "Group 2", "Group 3"]
        p = figure_factory.create_distplot(
            hist_data, group_labels, bin_size=[0.1, 0.25, 0.5]
        )

        # show the results
        event = plotly_chart(p)
        assert isinstance(event.figure, dict)



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional

from polyaxon.containers.contexts import CONTEXT_MOUNT_ARTIFACTS
from polyaxon.k8s import k8s_schemas
from polyaxon.polyflow import V1Init
from polyaxon.polypod.common import constants
from polyaxon.polypod.common.mounts import (
    get_artifacts_context_mount,
    get_connections_context_mount,
    get_mount_from_resource,
    get_mount_from_store,
)
from polyaxon.polypod.common.volumes import get_volume_name
from polyaxon.polypod.specs.contexts import PluginsContextsSpec
from polyaxon.schemas.types import V1ConnectionType, V1K8sResourceType
from polyaxon.utils.list_utils import to_list


def get_volume_mounts(
    contexts: PluginsContextsSpec,
    init: Optional[List[V1Init]],
    connections: Iterable[V1ConnectionType],
    secrets: Iterable[V1K8sResourceType],
    config_maps: Iterable[V1K8sResourceType] = None,
) -> List[k8s_schemas.V1VolumeMount]:
    init = init or []
    connections = connections or []
    secrets = secrets or []
    config_maps = config_maps or []

    volume_mounts = []
    volume_names = set()
    if contexts and contexts.collect_artifacts:
        volume_mounts += to_list(
            get_artifacts_context_mount(read_only=False), check_none=True
        )
        volume_names.add(constants.CONTEXT_VOLUME_ARTIFACTS)
    for init_connection in init:
        volume_name = (
            get_volume_name(init_connection.path)
            if init_connection.path
            else constants.CONTEXT_VOLUME_ARTIFACTS
        )
        mount_path = init_connection.path or CONTEXT_MOUNT_ARTIFACTS
        if volume_name in volume_names:
            continue
        volume_names.add(volume_name)
        volume_mounts += to_list(
            get_connections_context_mount(name=volume_name, mount_path=mount_path),
            check_none=True,
        )
    for store in connections:
        volume_mounts += to_list(get_mount_from_store(store=store), check_none=True)

    for secret in secrets:
        volume_mounts += to_list(
            get_mount_from_resource(resource=secret), check_none=True
        )

    for config_map in config_maps:
        volume_mounts += to_list(
            get_mount_from_resource(resource=config_map), check_none=True
        )

    return volume_mounts



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon import settings
from polyaxon.proxies.schemas.base import clean_config
from polyaxon.proxies.schemas.buffering import get_buffering_config
from polyaxon.proxies.schemas.charset import get_charset_config
from polyaxon.proxies.schemas.error_page import get_error_page_config
from polyaxon.proxies.schemas.gzip import get_gzip_config
from polyaxon.proxies.schemas.listen import get_listen_config
from polyaxon.proxies.schemas.locations import get_streams_locations_config
from polyaxon.proxies.schemas.logging import get_logging_config
from polyaxon.proxies.schemas.streams.gunicorn import get_gunicorn_config
from polyaxon.proxies.schemas.timeout import get_timeout_config


def get_base_config():
    config = [
        get_listen_config(is_proxy=False, port=settings.PROXIES_CONFIG.streams_port)
    ]
    config += [
        get_logging_config(),
        get_gzip_config(),
        get_charset_config(),
        get_buffering_config(),
        get_timeout_config(),
        get_gunicorn_config(),
        get_error_page_config(),
        get_streams_locations_config(),
    ]

    return clean_config(config)



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon import settings
from polyaxon.proxies.schemas.base import get_config

SSL_OPTIONS = """
# SSL
ssl_session_timeout 1d;
ssl_session_cache shared:SSL:50m;
ssl_session_tickets off;

# modern configuration
ssl_protocols TLSv1.2;
ssl_ciphers ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256;
ssl_prefer_server_ciphers on;

# OCSP Stapling
ssl_stapling on;
ssl_stapling_verify on;
resolver 1.1.1.1 1.0.0.1 8.8.8.8 8.8.4.4 208.67.222.222 208.67.220.220 valid=60s;
resolver_timeout 2s;

ssl_certificate      {ssl_path}/polyaxon.com.crt;
ssl_certificate_key  {ssl_path}/polyaxon.com.key;
"""  # noqa


def get_ssl_config():
    return get_config(
        options=SSL_OPTIONS, indent=0, ssl_path=settings.PROXIES_CONFIG.ssl_path
    )



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from polyaxon import settings
from polyaxon.logger import logger
from polyaxon.polyboard.logging import V1Log


class PolyaxonHandler(logging.Handler):
    def __init__(self, send_logs, **kwargs):
        self._send_logs = send_logs
        logging.Handler.__init__(
            self,
            level=kwargs.get(
                "level", settings.CLIENT_CONFIG.log_level or logging.NOTSET
            ),
        )

    def set_send_logs(self, send_logs):
        self._send_logs = send_logs

    def can_record(self, record):
        return not (
            record.name == "polyaxon"
            or record.name == "polyaxon.cli"
            or record.name.startswith("polyaxon")
        )

    def format_record(self, record):
        return V1Log(value=record.msg)

    def emit(self, record):  # pylint:disable=inconsistent-return-statements
        if settings.CLIENT_CONFIG.is_managed or not self.can_record(record):
            return
        try:
            return self._send_logs(self.format_record(record))
        except Exception:
            logger.warning("Polyaxon failed creating log record")



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from polyaxon.polyaxonfile.specs import kinds
from polyaxon.polyaxonfile.specs.base import BaseSpecification
from polyaxon.polyflow import V1CompiledOperation, V1Operation


class OperationSpecification(BaseSpecification):
    """The polyaxonfile specification for operations."""

    _SPEC_KIND = kinds.OPERATION

    CONFIG = V1Operation

    @classmethod
    def compile_operation(
        cls, config: V1Operation, override: Dict = None, override_post: bool = True
    ) -> V1CompiledOperation:
        return V1CompiledOperation.read(
            cls.generate_run_data(config, override, override_post)
        )

    @classmethod
    def generate_run_data(cls, config: V1Operation, override=None, override_post=True):
        op_config = config.to_light_dict()
        name = None
        if config.component:
            name = config.component.get_name()
        component_config = op_config.pop("component", {})
        if name:
            component_config["name"] = name
        values = [
            {"version": config.version},
            component_config,
            {"kind": kinds.COMPILED_OPERATION},
        ]
        op_override = {}
        for field in [
            cls.NAME,
            cls.DESCRIPTION,
            cls.TAGS,
            cls.PROFILE,
            cls.QUEUE,
            cls.CACHE,
            cls.PLUGINS,
            cls.TERMINATION,
            cls.MATRIX,
            cls.SCHEDULE,
            cls.DEPENDENCIES,
            cls.TRIGGER,
            cls.CONDITIONS,
            cls.SKIP_ON_UPSTREAM_SKIP,
        ]:
            override_field = op_config.get(field)
            if override_field:
                op_override[field] = override_field
        # Patch run
        run_patch = op_config.get(cls.RUN_PATCH)
        if run_patch:
            op_override[cls.RUN] = run_patch
        if override_post:
            if op_override:
                values.append(op_override)
            if override:
                values.append(override)
        else:
            if override:
                values.append(override)
            if op_override:
                values.append(op_override)
        return values



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.managers.project import ProjectManager


def _is_same_project(owner=None, project=None):
    local_project = ProjectManager.get_config()
    if project and project == local_project.name:
        return not all([owner, local_project.owner]) or owner == local_project.owner


def _cache_project(config, owner=None, project=None):
    if ProjectManager.is_initialized() and ProjectManager.is_locally_initialized():
        if _is_same_project(owner, project):
            ProjectManager.set_config(config)
            return

    ProjectManager.set_config(config, visibility=ProjectManager.VISIBILITY_GLOBAL)


def cache(config_manager, config, owner=None, project=None):
    if config_manager == ProjectManager:
        _cache_project(config=config, project=project, owner=owner)

    # Set caching only if we have an initialized project
    if not ProjectManager.is_initialized():
        return

    if not _is_same_project(owner, project):
        return

    visibility = (
        ProjectManager.VISIBILITY_LOCAL
        if ProjectManager.is_locally_initialized()
        else ProjectManager.VISIBILITY_GLOBAL
    )
    config_manager.set_config(config, visibility=visibility)



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from marshmallow import ValidationError
from tests.utils import BaseTestCase

from polyaxon import types
from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.polyflow import V1RunKind, dags
from polyaxon.polyflow.io import V1IO
from polyaxon.polyflow.operations import V1Operation
from polyaxon.polyflow.params import V1Param, ops_params
from polyaxon.polyflow.run import V1Dag


@pytest.mark.workflow_mark
class TestWorkflowV1Dags(BaseTestCase):
    def test_wrong_pipelines_ops(self):
        config_dict = {"operations": "foo"}
        with self.assertRaises(ValidationError):
            V1Dag.from_dict(config_dict)

        config_dict = {"operations": ["foo"]}
        with self.assertRaises(ValidationError):
            V1Dag.from_dict(config_dict)

    def test_dag_ops(self):
        config_dict = {
            "kind": "dag",
            "operations": [
                {
                    "name": "A",
                    "hubRef": "action1",
                    "description": "description A",
                    "tags": ["tag11", "tag12"],
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                    "termination": {"maxRetries": 2},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                    },
                },
                {
                    "urlRef": "https://url-to-temaplte.com",
                    "name": "B",
                    "description": "description B",
                    "tags": ["tag11", "tag12"],
                    "params": {
                        "param1": {"ref": "ops.A", "value": "outputs.x"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                    "termination": {"maxRetries": 2},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                    },
                },
                {
                    "dagRef": "my-template",
                    "name": "C",
                    "description": "description C",
                    "tags": ["tag31", "tag32"],
                    "params": {"param2": {"value": 12.34}, "param3": {"value": False}},
                    "termination": {"maxRetries": 5},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                            "restartPolicy": "Never",
                        },
                    },
                },
                {
                    "pathRef": "./relative/path/to/my-template.yaml",
                    "name": "D",
                    "description": "description D",
                    "tags": ["tag31", "tag32"],
                    "dependencies": ["B", "C"],
                    "termination": {"maxRetries": 3},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                            "restartPolicy": "Never",
                        },
                    },
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

    def test_pipelines_components(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "components": [
                {
                    "kind": "component",
                    "name": "experiment-template",
                    "description": "description experiment",
                    "tags": ["tag11", "tag12"],
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.FLOAT,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.S3,
                        }
                    ],
                    "termination": {"maxRetries": 2},
                    "run": {
                        "kind": V1RunKind.JOB,
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                            "restartPolicy": "Never",
                        },
                        "container": {
                            "resources": {"requests": {"cpu": "500m"}},
                            "image": "test",
                        },
                    },
                }
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        config.process_dag()
        config.validate_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

    def test_dag_job_component_with_correct_params(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "name": "experiment-template",
                    "params": {"input1": {"value": 1.1}, "input2": {"value": False}},
                    "termination": {"maxRetries": 2},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": "500m"}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                            "restartPolicy": "Never",
                        },
                    },
                    "component": {
                        "description": "description experiment",
                        "tags": ["tag11", "tag12"],
                        "inputs": [
                            {
                                "name": "input1",
                                "description": "some text",
                                "type": types.FLOAT,
                            },
                            {
                                "name": "input2",
                                "description": "some text",
                                "type": types.BOOL,
                                "isOptional": True,
                                "value": True,
                            },
                        ],
                        "outputs": [
                            {
                                "name": "output1",
                                "description": "some text",
                                "type": types.S3,
                            }
                        ],
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        config.process_dag()
        config.validate_dag()
        config.process_components(
            inputs=[V1IO.from_dict({"name": "input_pipe", "type": types.S3})]
        )

    def test_dag_structure(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "hubRef": "action1",
                    "name": "A",
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "urlRef": "https://url-to-temaplte.com",
                    "name": "B",
                    "params": {
                        "param1": {"value": "outputs.x", "ref": "ops.A"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "dagRef": "my-template",
                    "name": "C",
                    "params": {"param2": {"value": 12.34}, "param3": {"value": False}},
                },
                {
                    "pathRef": "./relative/path/to/my-template.yaml",
                    "name": "D",
                    "dependencies": ["B", "C"],
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A", "C"}
        assert dags.get_independent_ops(dag=dag) == {"A", "C"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert sorted_dag[0] in [["A", "C"], ["C", "A"]]
        assert sorted_dag[1] == ["B"]
        assert sorted_dag[2] == ["D"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "url_ref",
            "url": "https://url-to-temaplte.com",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "my-template",
        }
        assert config.dag["C"].upstream == set()
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "path_ref",
            "path": "./relative/path/to/my-template.yaml",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

    def test_dag_matrix_ops(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A"},
                {"hubRef": "event1", "name": "B"},
                {"dagRef": "foo", "name": "C"},
                {"dagRef": "bar", "name": "D"},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A", "B", "C", "D"}
        assert dags.get_independent_ops(dag=dag) == {"A", "B", "C", "D"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        assert len(config.sort_topologically(dag=dag)) == 1  # order can be any
        assert len(config.sort_topologically(dag=dag)[0]) == 4  # order can be any

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == set()

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == set()
        assert config.dag["B"].downstream == set()

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == set()
        assert config.dag["C"].downstream == set()

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "bar",
        }
        assert config.dag["D"].upstream == set()
        assert config.dag["D"].downstream == set()

    def test_dag_sequential_ops(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A"},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["B"]},
                {"dagRef": "bar", "name": "D", "dependencies": ["C"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]
        assert dags.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"C"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"B"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "bar",
        }
        assert config.dag["D"].upstream == {"C"}
        assert config.dag["D"].downstream == set()

    def test_dag_all_downstream(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "hubRef": "action1",
                    "name": "A",
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                    "dependencies": ["B", "C", "D"],
                },
                {"hubRef": "event1", "name": "B"},
                {"hubRef": "event2", "name": "C"},
                {"hubRef": "event3", "name": "D"},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        # No components
        config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"B", "C", "D"}
        assert dags.get_independent_ops(dag=dag) == {"B", "C", "D"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])

        sorted_dag = config.sort_topologically(dag=dag)
        assert dags.sort_topologically(dag=dag) == sorted_dag
        # Sort is not guaranteed at stage 0
        assert len(sorted_dag[0]) == 3
        assert set(sorted_dag[0]) == {"B", "C", "D"}
        assert sorted_dag[1] == ["A"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == {"B", "C", "D"}
        assert config.dag["A"].downstream == set()

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == set()
        assert config.dag["B"].downstream == {"A"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event2",
        }
        assert config.dag["C"].upstream == set()
        assert config.dag["C"].downstream == {"A"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event3",
        }
        assert config.dag["D"].upstream == set()
        assert config.dag["D"].downstream == {"A"}

    def test_dag_acyclic_deps(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A", "dependencies": ["B"]},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["A"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == set()
        assert dags.get_independent_ops(dag=dag) == set()
        assert config.get_orphan_ops(dag=dag) == set()
        assert dags.get_orphan_ops(dag=dag) == set()
        with self.assertRaises(PolyaxonSchemaError):
            config.sort_topologically(dag=dag)
        with self.assertRaises(PolyaxonSchemaError):
            dags.sort_topologically(dag=dag)

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == {"B"}
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"A"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == set()

    def test_dag_circular_deps(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A", "dependencies": ["C"]},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["B"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == set()
        assert dags.get_independent_ops(dag=dag) == set()
        assert config.get_orphan_ops(dag=dag) == set()
        assert dags.get_orphan_ops(dag=dag) == set()
        with self.assertRaises(PolyaxonSchemaError):
            config.sort_topologically(dag=dag)
        with self.assertRaises(PolyaxonSchemaError):
            dags.sort_topologically(dag=dag)

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == {"C"}
        assert config.dag["A"].downstream == {"B"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"C"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"B"}
        assert config.dag["C"].downstream == {"A"}

    def test_dag_adding_ops_one_by_one_manually(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A"},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["A"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert dags.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]  # Only this one has consistent order
        assert sorted_dag[1] in [["B", "C"], ["C", "B"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == set()

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == set()

        operation_d = V1Operation(hub_ref="action4", name="D", dependencies=["B", "C"])
        config.add_op(operation_d)

        operation_e = V1Operation(hub_ref="action4", name="E", dependencies=["A"])
        config.add_op(operation_e)

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {"B", "C", "E"}
        assert sorted_dag[2] == ["D"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C", "E"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

        assert config.dag["E"].op.name == "E"
        assert config.dag["E"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["E"].upstream == {"A"}
        assert config.dag["E"].downstream == set()

        # Adding the same ops should not alter the behavior
        config.add_ops([operation_d, operation_e])

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {"B", "C", "E"}
        assert sorted_dag[2] == ["D"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C", "E"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

        assert config.dag["E"].op.name == "E"
        assert config.dag["E"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["E"].upstream == {"A"}
        assert config.dag["E"].downstream == set()

    def test_dag_adding_ops_many_manually(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A"},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["A"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]  # Only this one has consistent order
        assert sorted_dag[1] in [["B", "C"], ["C", "B"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == set()

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == set()

        operation_d = V1Operation(hub_ref="action4", name="D", dependencies=["B", "C"])
        operation_e = V1Operation(hub_ref="action4", name="E", dependencies=["A"])
        config.add_ops([operation_d, operation_e])

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        assert len(config.dag) == 5
        assert config.get_independent_ops(dag=config.dag) == {"A"}
        assert dags.get_independent_ops(dag=config.dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {"B", "C", "E"}
        assert sorted_dag[2] == ["D"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C", "E"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

        assert config.dag["E"].op.name == "E"
        assert config.dag["E"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["E"].upstream == {"A"}
        assert config.dag["E"].downstream == set()

        # Adding the same ops should not alter the behavior
        config.add_op(operation_d)
        config.add_op(operation_e)

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 5
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        # Sort is not guaranteed at stage 1
        assert len(sorted_dag[1]) == 3
        assert set(sorted_dag[1]) == {"B", "C", "E"}
        assert sorted_dag[2] == ["D"]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C", "E"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

        assert config.dag["E"].op.name == "E"
        assert config.dag["E"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action4",
        }
        assert config.dag["E"].upstream == {"A"}
        assert config.dag["E"].downstream == set()

    def test_dag_dependency_from_params(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "hubRef": "echo",
                    "name": "A",
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "hubRef": "echo",
                    "name": "B",
                    "params": {
                        "param1": {"ref": "ops.A", "value": "outputs.x"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "hubRef": "echo",
                    "name": "C",
                    "params": {
                        "param1": {"ref": "ops.A", "value": "outputs.x"},
                        "param2": {"ref": "ops.B", "value": "outputs.x"},
                    },
                },
                {"hubRef": "echo", "name": "D", "dependencies": ["B", "C"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        # Todo must resolve the components from the hub
        config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]
        assert dags.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"C", "D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["C"].upstream == {"A", "B"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

    def test_dag_dependency_from_metainfo(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "hubRef": "echo",
                    "name": "A",
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "hubRef": "echo",
                    "name": "B",
                    "params": {
                        "param1": {"value": "status", "ref": "ops.A"},
                        "param2": {"value": "project_name", "ref": "ops.A"},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "hubRef": "echo",
                    "name": "C",
                    "params": {
                        "param1": {"value": "inputs", "ref": "ops.A"},
                        "param2": {"value": "uuid", "ref": "ops.B"},
                        "param3": {"value": "outputs", "ref": "ops.B"},
                    },
                },
                {"hubRef": "echo", "name": "D", "dependencies": ["B", "C"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        # Todo must resolve the components from the hub
        config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]
        assert dags.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"C", "D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["C"].upstream == {"A", "B"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

    def test_dag_dependency_and_params(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "hubRef": "echo",
                    "name": "A",
                    "params": {
                        "param1": {"value": "text"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
                {
                    "hubRef": "echo",
                    "name": "B",
                    "params": {
                        "param1": {"value": "outputs.x", "ref": "ops.A"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                    "dependencies": ["A"],
                },
                {
                    "hubRef": "echo",
                    "name": "C",
                    "params": {"param2": {"value": "outputs.x", "ref": "ops.B"}},
                    "dependencies": ["A"],
                },
                {"hubRef": "echo", "name": "D", "dependencies": ["B", "C"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        # TODO resolve components from hub
        config.process_components()
        config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        assert config.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]
        assert dags.sort_topologically(dag=dag) == [["A"], ["B"], ["C"], ["D"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == {"C", "D"}

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["C"].upstream == {"A", "B"}
        assert config.dag["C"].downstream == {"D"}

        assert config.dag["D"].op.name == "D"
        assert config.dag["D"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "echo",
        }
        assert config.dag["D"].upstream == {"B", "C"}
        assert config.dag["D"].downstream == set()

    def test_dag_orphan_ops(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"hubRef": "action1", "name": "A"},
                {"hubRef": "event1", "name": "B", "dependencies": ["A"]},
                {"dagRef": "foo", "name": "C", "dependencies": ["A", "E"]},
            ],
        }

        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        assert config.dag == {}

        # Process the dag
        config.process_dag()
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()
        with self.assertRaises(PolyaxonSchemaError):
            config.validate_dag()
        dag = config.dag
        assert len(dag) == 4
        assert config.get_independent_ops(dag=dag) == {"A", "E"}
        assert dags.get_independent_ops(dag=dag) == {"A", "E"}
        assert config.get_orphan_ops(dag=dag) == {"E"}
        assert dags.get_orphan_ops(dag=dag) == {"E"}
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] in [["A", "E"], ["E", "A"]]
        assert sorted_dag[1] in [["B", "C"], ["C", "B"]]

        assert config.dag["A"].op.name == "A"
        assert config.dag["A"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "action1",
        }
        assert config.dag["A"].upstream == set()
        assert config.dag["A"].downstream == {"B", "C"}

        assert config.dag["B"].op.name == "B"
        assert config.dag["B"].op.template.to_dict() == {
            "kind": "hub_ref",
            "name": "event1",
        }
        assert config.dag["B"].upstream == {"A"}
        assert config.dag["B"].downstream == set()

        assert config.dag["C"].op.name == "C"
        assert config.dag["C"].op.template.to_dict() == {
            "kind": "dag_ref",
            "name": "foo",
        }
        assert config.dag["C"].upstream == {"A", "E"}
        assert config.dag["C"].downstream == set()

    def test_dag_with_duplicate_job_names(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"dagRef": "build-template1", "name": "A"},
                {"dagRef": "build-template1", "name": "A"},
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "build-template",
                    "description": "description build",
                    "tags": ["tag11", "tag12"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "job-template",
                    "description": "description build",
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

    def test_dag_with_op_requesting_undefined_template(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"dagRef": "build-template1", "name": "A"},
                {
                    "dagRef": "build-template1",
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "connections": ["data2", "data2"],
                    },
                    "name": "B",
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "build-template2",
                    "description": "description build",
                    "tags": ["kaniko"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(PolyaxonSchemaError):
            config.process_components()

    def test_dag_with_template_not_defining_inputs(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "build-template",
                    "name": "A",
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "connections": ["foo", "boo"],
                        "container": {"args": ["--branch=dev"]},
                    },
                },
                {"dagRef": "job-template", "name": "B", "dependencies": ["A"]},
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "build-template",
                    "tags": ["kaniko"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        # Trying to set op template before processing components
        with self.assertRaises(PolyaxonSchemaError):
            config.set_op_component("A")
        config.process_components()
        # Trying to set op template before processing dag
        with self.assertRaises(PolyaxonSchemaError):
            config.set_op_component("A")
        config.process_dag()
        assert config.dag["A"].op.has_component_reference is False
        assert config.dag["B"].op.has_component_reference is False
        config.set_op_component("A")
        assert config.dag["B"].op.has_component_reference is False
        assert config.dag["A"].op.has_component_reference is True
        assert (
            config.dag["A"].op.template == config._components_by_names["build-template"]
        )
        config.set_op_component("B")
        assert config.dag["B"].op.has_component_reference is True
        assert (
            config.dag["B"].op.template == config._components_by_names["job-template"]
        )

    def test_dag_with_template_not_defining_inputs_and_ops_with_params(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "name": "A",
                    "component": {
                        "name": "build-template",
                        "tags": ["kaniko"],
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                },
                {
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "param1": {"value": "outputs.x", "ref": "ops.A"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                    "component": {
                        "name": "job-template",
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_pipelines_with_template_not_defining_inputs_and_ops_with_params_template(
        self,
    ):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"dagRef": "build-template", "name": "A"},
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "param1": {"value": "outputs.x", "ref": "ops.A"},
                        "param2": {"value": 12},
                        "param3": {"value": "https://foo.com"},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "build-template",
                    "tags": ["kaniko"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_ops_template_required_inputs(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "name": "A",
                    "component": {
                        "name": "job-template",
                        "inputs": [
                            {
                                "name": "input1",
                                "description": "some text",
                                "type": types.FLOAT,
                            }
                        ],
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_pipelines_with_ops_template_required_inputs_template(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [{"dagRef": "job-template", "name": "A"}],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.FLOAT,
                        }
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_ops_template_optional_inputs(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "name": "A",
                    "component": {
                        "name": "job-template",
                        "inputs": [
                            {
                                "name": "input1",
                                "description": "some text",
                                "type": types.FLOAT,
                                "isOptional": True,
                                "value": 12.2,
                            }
                        ],
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        config.process_components()

    def test_pipelines_with_ops_template_optional_inputs_template(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [{"dagRef": "job-template", "name": "A"}],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.FLOAT,
                            "isOptional": True,
                            "value": 12.2,
                        }
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        config.process_components()

    def test_dag_with_ops_template_optional_inputs_and_wrong_param(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "name": "A",
                    "params": {"input1": {"value": "foo"}},
                    "component": {
                        "name": "job-template",
                        "inputs": [
                            {
                                "name": "input1",
                                "description": "some text",
                                "type": types.FLOAT,
                                "isOptional": True,
                                "value": 12.2,
                            }
                        ],
                        "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                    },
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_pipelines_with_ops_template_optional_inputs_and_wrong_param_components(
        self,
    ):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "job-template",
                    "name": "A",
                    "params": {"input1": {"value": "foo"}},
                }
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.FLOAT,
                            "isOptional": True,
                            "value": 12.2,
                        }
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        config_to_light = config.to_light_dict()
        assert config_to_light == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_ops_template_validation(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "job-template",
                    "name": "A",
                    "params": {
                        "input1": {"value": "sdf"},
                        "input2": {"value": 12.0},
                        "input3": {"value": False},
                    },
                },
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": "ooo"},
                        "input2": {"value": 12.123},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.STR,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.FLOAT,
                        },
                        {
                            "name": "input3",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict

        # Trying to set op template before processing components
        with self.assertRaises(PolyaxonSchemaError):
            config.set_op_component("A")
        config.process_components()
        # Trying to set op template before processing dag
        with self.assertRaises(PolyaxonSchemaError):
            config.set_op_component("A")
        config.process_dag()
        assert config.dag["A"].op.has_component_reference is False
        assert config.dag["B"].op.has_component_reference is False
        config.set_op_component("A")
        assert config.dag["B"].op.has_component_reference is False
        assert config.dag["A"].op.has_component_reference is True
        assert (
            config.dag["A"].op.template == config._components_by_names["job-template"]
        )
        config.set_op_component("B")
        assert config.dag["B"].op.has_component_reference is True
        assert (
            config.dag["B"].op.template == config._components_by_names["job-template"]
        )

    def test_dag_with_wrong_refs(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "job-template",
                    "name": "A",
                    "params": {
                        "input1": {"value": "sdf"},
                        "input2": {"value": 12.0},
                        "input3": {"value": False},
                    },
                },
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": "outputs.output1", "ref": "ops.A"},
                        "input2": {"value": 12.123},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.STR,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.FLOAT,
                        },
                        {
                            "name": "input3",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_correct_refs(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "job-template",
                    "name": "A",
                    "params": {
                        "input1": {"value": 2},
                        "input2": {"value": "gs://bucket/path/to/blob/"},
                        "input4": {"value": "failed"},
                        "input5": {"value": {"foo": "bar"}},
                    },
                },
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": "outputs.output1", "ref": "ops.A"},
                        "input2": {"value": "gs://bucket/path/to/blob/"},
                        "input4": {"value": "status", "ref": "ops.A"},
                        "input5": {"value": "inputs", "ref": "ops.A"},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.INT,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.GCS,
                        },
                        {
                            "name": "input3",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                        {"name": "input4", "description": "status", "type": types.STR},
                        {"name": "input5", "description": "dict", "type": types.DICT},
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.INT,
                            "isOptional": True,
                            "value": 123,
                        }
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_components()

    def test_dag_with_correct_ref_and_wrong_ref_type(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "job-template",
                    "name": "A",
                    "params": {
                        "input1": {"value": 2},
                        "input2": {"value": "gs://bucket/path/to/blob/"},
                        "output1": {"value": 123},
                    },
                },
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": 3},
                        "input2": {"value": "outputs.output1", "ref": "ops.A"},
                        "output1": {"value": 123},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.INT,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.GCS,
                        },
                        {
                            "name": "input3",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.INT,
                        }
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                }
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_template_not_defining_inputs_and_ops_refs_params(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {"dagRef": "build-template", "name": "A"},
                {
                    "dagRef": "job-template",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {"param1": {"value": "outputs.x", "ref": "ops.A"}},
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "job-template",
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "build-template",
                    "tags": ["kaniko"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
            ],
        }
        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.process_components()

    def test_dag_with_ops_and_components(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "build-template",
                    "name": "A",
                    "description": "description A",
                    "tags": ["tag11", "tag12"],
                    "termination": {"maxRetries": 2},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                    },
                },
                {
                    "dagRef": "experiment-template",
                    "name": "B",
                    "description": "description B",
                    "tags": ["tag21", "tag22"],
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": 11.1},
                        "input2": {"value": False},
                        "input3": {
                            "value": "outputs.foo",
                            "ref": "runs.64332180bfce46eba80a65caf73c5396",
                        },
                        "output1": {"value": "S3://foo.com"},
                    },
                    "termination": {"maxRetries": 3},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                    },
                },
                {
                    "dagRef": "group-template",
                    "name": "C",
                    "description": "description C",
                    "tags": ["tag31", "tag32"],
                    "params": {
                        "input1": {"value": "outputs.output1", "ref": "ops.B"},
                        "input2": {"value": "outputs.output2", "ref": "ops.B"},
                        "output1": {"value": "S3://foo.com"},
                    },
                    "termination": {"maxRetries": 5},
                    "runPatch": {
                        "kind": V1RunKind.JOB,
                        "container": {"resources": {"requests": {"cpu": 1}}},
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "experiment-template",
                    "description": "description experiment",
                    "tags": ["tag11", "tag12"],
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.FLOAT,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                        {
                            "name": "input3",
                            "description": "some text",
                            "type": types.INT,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.S3,
                        },
                        {
                            "name": "output2",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "termination": {"maxRetries": 2},
                    "run": {
                        "kind": V1RunKind.JOB,
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                        "container": {
                            "image": "test",
                            "resources": {"requests": {"cpu": 1}},
                        },
                    },
                },
                {
                    "kind": "component",
                    "name": "group-template",
                    "description": "description group",
                    "tags": ["tag11", "tag12"],
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.S3,
                        },
                        {
                            "name": "input2",
                            "description": "some text",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.S3,
                        }
                    ],
                    "termination": {"maxRetries": 2},
                    "run": {
                        "kind": V1RunKind.JOB,
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                        "container": {
                            "image": "test",
                            "resources": {"requests": {"cpu": 1}},
                        },
                    },
                },
                {
                    "kind": "component",
                    "name": "build-template",
                    "description": "description build",
                    "tags": ["tag11", "tag12"],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "build-template2",
                    "description": "description build",
                    "tags": ["tag11", "tag12", "kaniko"],
                    "termination": {"maxRetries": 2},
                    "run": {
                        "kind": V1RunKind.JOB,
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                        "container": {
                            "image": "test",
                            "resources": {"requests": {"cpu": 1}},
                        },
                    },
                },
                {
                    "kind": "component",
                    "name": "job-template",
                    "description": "description job",
                    "tags": ["tag11", "tag12"],
                    "inputs": [
                        {
                            "name": "input1",
                            "description": "some text",
                            "type": types.S3,
                            "isOptional": True,
                            "value": "s3://foo",
                        }
                    ],
                    "outputs": [
                        {
                            "name": "output1",
                            "description": "some text",
                            "type": types.S3,
                        }
                    ],
                    "termination": {"maxRetries": 2},
                    "run": {
                        "kind": V1RunKind.JOB,
                        "environment": {
                            "nodeSelector": {"polyaxon": "core"},
                            "serviceAccountName": "service",
                            "imagePullSecrets": ["secret1", "secret2"],
                        },
                        "container": {
                            "image": "test",
                            "resources": {"requests": {"cpu": 1}},
                        },
                    },
                },
            ],
        }

        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        config.process_components()
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        assert sorted_dag[1] == ["B"]
        assert sorted_dag[2] == ["C"]

        # op upstreams
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["A"].op.params
        )
        assert op_upstream_by_names == {}
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["B"].op.params
        )
        assert op_upstream_by_names == {}
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["C"].op.params
        )
        assert len(op_upstream_by_names["B"]) == 2
        if op_upstream_by_names["B"][0].name == "input1":
            assert op_upstream_by_names["B"][0] == ops_params.ParamSpec(
                name="input1",
                iotype=None,
                param=V1Param(ref="ops.B", value="outputs.output1"),
                is_flag=None,
                is_list=None,
            )
            assert op_upstream_by_names["B"][1] == ops_params.ParamSpec(
                name="input2",
                iotype=None,
                param=V1Param(ref="ops.B", value="outputs.output2"),
                is_flag=None,
                is_list=None,
            )

        else:
            assert op_upstream_by_names["B"][1] == ops_params.ParamSpec(
                name="input1",
                iotype=None,
                param=V1Param(ref="ops.B", value="outputs.output1"),
                is_flag=None,
                is_list=None,
            )
            assert op_upstream_by_names["B"][0] == ops_params.ParamSpec(
                name="input2",
                iotype=None,
                param=V1Param(ref="ops.B", value="outputs.output2"),
                is_flag=None,
                is_list=None,
            )

        # run upstreams
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["A"].op.params
        )
        assert run_upstream_by_names == {}
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["B"].op.params
        )
        assert run_upstream_by_names["64332180bfce46eba80a65caf73c5396"] == [
            ops_params.ParamSpec(
                name="input3",
                iotype=None,
                param=V1Param(
                    ref="runs.64332180bfce46eba80a65caf73c5396", value="outputs.foo"
                ),
                is_flag=None,
                is_list=None,
            )
        ]
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["C"].op.params
        )
        assert run_upstream_by_names == {}

        # pipeline upstreams
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["A"].op.params
        )
        assert pipeline_by_names == {}
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["B"].op.params
        )
        assert pipeline_by_names == {}
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["C"].op.params
        )
        assert pipeline_by_names == {}

    def test_pipeline_context(self):
        config_dict = {
            "kind": V1RunKind.DAG,
            "operations": [
                {
                    "dagRef": "A",
                    "name": "A",
                    "params": {
                        "input1": {"value": 11.1},
                        "input2": {"value": False},
                        "input3": {
                            "value": "outputs.foo",
                            "ref": "runs.64332180bfce46eba80a65caf73c5396",
                        },
                        "input4": {"value": "s3://foo"},
                    },
                },
                {
                    "dagRef": "B",
                    "name": "B",
                    "dependencies": ["A"],
                    "params": {
                        "input1": {"value": "inputs.input4", "ref": "ops.A"},
                        "input2": {"value": "outputs.output1", "ref": "ops.A"},
                        "input3": {"value": "status", "ref": "ops.A"},
                        "input4": {"value": "inputs", "ref": "ops.A"},
                    },
                },
                {
                    "dagRef": "B",
                    "name": "C",
                    "params": {
                        "input1": {"ref": "dag", "value": "inputs.input_pipe"},
                        "input2": {"ref": "ops.B", "value": "outputs.output1"},
                        "input3": {"ref": "ops.A", "value": "status"},
                        "input4": {"ref": "ops.B", "value": "inputs"},
                    },
                },
            ],
            "components": [
                {
                    "kind": "component",
                    "name": "A",
                    "inputs": [
                        {"name": "input1", "type": types.FLOAT},
                        {
                            "name": "input2",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                        {
                            "name": "input3",
                            "type": types.INT,
                            "isOptional": True,
                            "value": True,
                        },
                        {"name": "input4", "type": types.S3},
                    ],
                    "outputs": [
                        {"name": "output1", "type": types.S3},
                        {
                            "name": "output2",
                            "type": types.BOOL,
                            "isOptional": True,
                            "value": True,
                        },
                    ],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
                {
                    "kind": "component",
                    "name": "B",
                    "inputs": [
                        {"name": "input1", "type": types.S3},
                        {"name": "input2", "type": types.S3},
                        {"name": "input3", "type": types.STR},
                        {"name": "input4", "type": types.DICT},
                    ],
                    "outputs": [{"name": "output1", "type": types.S3}],
                    "run": {"kind": V1RunKind.JOB, "container": {"image": "test"}},
                },
            ],
        }

        config = V1Dag.from_dict(config_dict)
        assert config.to_light_dict() == config_dict
        config.process_dag()
        config.validate_dag()
        config.process_components(
            inputs=[V1IO.from_dict({"name": "input_pipe", "type": types.S3})]
        )
        dag = config.dag
        assert len(dag) == 3
        assert config.get_independent_ops(dag=dag) == {"A"}
        assert dags.get_independent_ops(dag=dag) == {"A"}
        assert config.get_orphan_ops(dag=dag) == set([])
        assert dags.get_orphan_ops(dag=dag) == set([])
        sorted_dag = config.sort_topologically(dag=dag)
        assert config.sort_topologically(dag=dag) == sorted_dag
        assert sorted_dag[0] == ["A"]
        assert sorted_dag[1] == ["B"]
        assert sorted_dag[2] == ["C"]

        # op upstreams
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["A"].op.params
        )
        assert op_upstream_by_names == {}
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["B"].op.params
        )

        if op_upstream_by_names["A"][0].name == "input1":
            assert op_upstream_by_names["A"][0] == ops_params.ParamSpec(
                name="input1",
                iotype=None,
                param=V1Param(ref="ops.A", value="inputs.input4"),
                is_flag=None,
                is_list=None,
            )
            assert op_upstream_by_names["A"][1] == ops_params.ParamSpec(
                name="input2",
                iotype=None,
                param=V1Param(ref="ops.A", value="outputs.output1"),
                is_flag=None,
                is_list=None,
            )

        else:
            assert op_upstream_by_names["A"][1] == ops_params.ParamSpec(
                name="input1",
                iotype=None,
                param=V1Param(ref="ops.A", value="inputs.input4"),
                is_flag=None,
                is_list=None,
            )
            assert op_upstream_by_names["A"][0] == ops_params.ParamSpec(
                name="input2",
                iotype=None,
                param=V1Param(ref="ops.A", value="outputs.output1"),
                is_flag=None,
                is_list=None,
            )
        op_upstream_by_names = ops_params.get_upstream_op_params_by_names(
            params=config.dag["C"].op.params
        )
        assert op_upstream_by_names == {
            "A": [
                ops_params.ParamSpec(
                    name="input3",
                    iotype=None,
                    param=V1Param(ref="ops.A", value="status"),
                    is_flag=None,
                    is_list=None,
                )
            ],
            "B": [
                ops_params.ParamSpec(
                    name="input2",
                    iotype=None,
                    param=V1Param(ref="ops.B", value="outputs.output1"),
                    is_flag=None,
                    is_list=None,
                ),
                ops_params.ParamSpec(
                    name="input4",
                    iotype=None,
                    param=V1Param(ref="ops.B", value="inputs"),
                    is_flag=None,
                    is_list=None,
                ),
            ],
        }

        # run upstreams
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["A"].op.params
        )
        assert run_upstream_by_names["64332180bfce46eba80a65caf73c5396"] == [
            ops_params.ParamSpec(
                name="input3",
                iotype=None,
                param=V1Param(
                    ref="runs.64332180bfce46eba80a65caf73c5396", value="outputs.foo"
                ),
                is_flag=None,
                is_list=None,
            )
        ]
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["B"].op.params
        )
        assert run_upstream_by_names == {}
        run_upstream_by_names = ops_params.get_upstream_run_params_by_names(
            params=config.dag["C"].op.params
        )
        assert run_upstream_by_names == {}

        # pipeline upstreams
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["A"].op.params
        )
        assert pipeline_by_names == {}
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["B"].op.params
        )
        assert pipeline_by_names == {}
        pipeline_by_names = ops_params.get_dag_params_by_names(
            params=config.dag["C"].op.params
        )
        assert pipeline_by_names["_"] == [
            ops_params.ParamSpec(
                name="input1",
                iotype=None,
                param=V1Param(ref="dag", value="inputs.input_pipe"),
                is_flag=None,
                is_list=None,
            )
        ]



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile

import pytest

from tests.utils import BaseTestCase

from polyaxon.polyaxonfile import (
    DEFAULT_POLYAXON_FILE_EXTENSION,
    DEFAULT_POLYAXON_FILE_NAME,
    check_default_path,
)


@pytest.mark.polyaxonfile_mark
class TestDefaultFile(BaseTestCase):
    def test_default_not_found(self):
        path = tempfile.mkdtemp()
        assert check_default_path(path=path) is None

    def test_polyaxon_found(self):
        def create_file(path, filename, ext):
            fpath = "{}/{}.{}".format(path, filename, ext)
            open(fpath, "w")

        for filename in DEFAULT_POLYAXON_FILE_NAME:
            for ext in DEFAULT_POLYAXON_FILE_EXTENSION:
                path = tempfile.mkdtemp()
                create_file(path, filename, ext)
                assert check_default_path(path=path)



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml

from typing import TextIO, Union

try:
    from yaml import CSafeLoader as _SafeLoader, CSafeDumper as _SafeDumper
except ImportError:
    from yaml import SafeLoader as _SafeLoader, SafeDumper as _SafeDumper


def dump(data, stream=None):
    return yaml.dump(
        data,
        stream=stream,
        Dumper=_SafeDumper,
        default_flow_style=False,
        allow_unicode=True,
    )


def safe_load(filepath: Union[str, TextIO]):
    return yaml.load(filepath, Loader=_SafeLoader)



#!/usr/bin/python
#
# Copyright 2018-2020 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

import mock

from tests.test_transports.utils import BaseTestCaseTransport

from polyaxon.client.transport import Transport
from polyaxon.client.transport.utils import Bar


class TestHttpTransport(BaseTestCaseTransport):
    # pylint:disable=protected-access
    def setUp(self):
        super().setUp()
        self.transport = Transport()

    def test_create_progress_callback(self):
        encoder = mock.MagicMock()
        encoder.configure_mock(len=10)
        _, progress_bar = self.transport.create_progress_callback(encoder)
        assert isinstance(progress_bar, Bar)

    def test_format_sizeof(self):
        assert self.transport.format_sizeof(10) == "10.0B"
        assert self.transport.format_sizeof(10000) == "9.8KiB"
        assert self.transport.format_sizeof(100000) == "97.7KiB"
        assert self.transport.format_sizeof(10000000) == "9.5MiB"
        assert self.transport.format_sizeof(10000000000) == "9.3GiB"

    def test_session(self):
        assert hasattr(self.transport, "_session") is False
        assert isinstance(self.transport.session, requests.Session)
        assert isinstance(self.transport._session, requests.Session)
